@@ -1,13 +1,26 @@
 /* -*- Mode: C; tab-width: 8; indent-tabs-mode: t; c-basic-offset: 8 -*- */
-/*
- * e-tree.c: A graphical view of a tree.
+/* 
+ * e-tree.c
+ * Copyright 2000, 2001, Ximian, Inc.
  *
- * Author:
- *   Miguel de Icaza (miguel@ximian.com)
+ * Authors:
  *   Chris Lahey <clahey@ximian.com>
  *
- * Copyright 1999, 2000, 2001, Ximian, Inc
+ * This library is free software; you can redistribute it and/or
+ * modify it under the terms of the GNU Library General Public
+ * License, version 2, as published by the Free Software Foundation.
+ *
+ * This library is distributed in the hope that it will be useful, but
+ * WITHOUT ANY WARRANTY; without even the implied warranty of
+ * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
+ * Library General Public License for more details.
+ *
+ * You should have received a copy of the GNU Library General Public
+ * License along with this library; if not, write to the Free Software
+ * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
+ * 02111-1307, USA.
  */
+
 #include <config.h>
 #include <stdlib.h>
 #include <stdio.h>
@@ -45,6 +58,14 @@
 
 static GtkObjectClass *parent_class;
 
+#define d(x)
+
+#if d(!)0
+#define e_table_item_leave_edit_(x) (e_table_item_leave_edit((x)), g_print ("%s: e_table_item_leave_edit\n", __FUNCTION__))
+#else
+#define e_table_item_leave_edit_(x) (e_table_item_leave_edit((x)))
+#endif
+
 enum {
 	CURSOR_CHANGE,
 	CURSOR_ACTIVATED,
@@ -53,6 +74,7 @@ enum {
 	RIGHT_CLICK,
 	CLICK,
 	KEY_PRESS,
+	START_DRAG,
 
 	TREE_DRAG_BEGIN,
 	TREE_DRAG_END,
@@ -73,7 +95,8 @@ enum {
 	ARG_HORIZONTAL_DRAW_GRID,
 	ARG_VERTICAL_DRAW_GRID,
 	ARG_DRAW_FOCUS,
-	ARG_ETTA
+	ARG_ETTA,
+	ARG_UNIFORM_ROW_HEIGHT,
 };
 
 struct ETreePriv {
@@ -90,6 +113,8 @@ struct ETreePriv {
 	ETableSpecification *spec;
 
 	int reflow_idle_id;
+	int scroll_idle_id;
+	int hover_idle_id;
 
 	int table_model_change_id;
 	int table_row_change_id;
@@ -117,19 +142,31 @@ struct ETreePriv {
 
 	guint horizontal_scrolling : 1;
 
+	guint scroll_down : 1;
+
+	guint do_drag : 1;
+
+	guint uniform_row_height : 1;
+
 	ECursorMode cursor_mode;
 
 	int drop_row;
 	ETreePath drop_path;
 	int drop_col;
 
+	GnomeCanvasItem *drop_highlight;
+	int last_drop_x;
+	int last_drop_y;
+	int last_drop_time;
+	GdkDragContext *last_drop_context;
+
+	int hover_x;
+	int hover_y;
+
 	int drag_row;
 	ETreePath drag_path;
 	int drag_col;
 	ETreeDragSourceSite *site;
-	
-	int drag_source_button_press_event_id;
-	int drag_source_motion_notify_event_id;
 };
 
 static gint et_signals [LAST_SIGNAL] = { 0, };
@@ -176,12 +213,14 @@ static void et_drag_data_received(GtkWidget *widget,
 				  guint info,
 				  guint time,
 				  ETree *et);
-static gint e_tree_drag_source_event_cb (GtkWidget      *widget,
-					 GdkEvent       *event,
-					 ETree         *tree);
 
 static gint et_focus (GtkContainer *container, GtkDirectionType direction);
 
+static void scroll_off (ETree *et);
+static void scroll_on (ETree *et, gboolean down);
+static void hover_off (ETree *et);
+static void hover_on (ETree *et, int x, int y);
+
 static void
 et_disconnect_from_etta (ETree *et)
 {
@@ -217,6 +256,9 @@ et_destroy (GtkObject *object)
 		g_source_remove(et->priv->reflow_idle_id);
 	et->priv->reflow_idle_id = 0;
 
+	scroll_off (et);
+	hover_off (et);
+
 	et_disconnect_from_etta (et);
 
 	gtk_object_unref (GTK_OBJECT (et->priv->etta));
@@ -241,6 +283,7 @@ et_destroy (GtkObject *object)
 	gtk_widget_destroy (GTK_WIDGET (et->priv->table_canvas));
 
 	g_free(et->priv);
+	et->priv = NULL;
 
 	(*parent_class->destroy)(object);
 }
@@ -248,8 +291,8 @@ et_destroy (GtkObject *object)
 static void
 e_tree_init (GtkObject *object)
 {
-	ETree *e_tree = E_TREE (object);
-	GtkTable *gtk_table = GTK_TABLE (object);
+	ETree *e_tree                                    = E_TREE (object);
+	GtkTable *gtk_table                              = GTK_TABLE (object);
 
 	GTK_WIDGET_SET_FLAGS (e_tree, GTK_CAN_FOCUS);
 
@@ -267,13 +310,16 @@ e_tree_init (GtkObject *object)
 	e_tree->priv->sort_info                          = NULL;
 	e_tree->priv->sorter                             = NULL;
 	e_tree->priv->reflow_idle_id                     = 0;
+	e_tree->priv->scroll_idle_id                     = 0;
+	e_tree->priv->hover_idle_id                     = 0;
 
-	e_tree->priv->alternating_row_colors		 = 1;
+	e_tree->priv->alternating_row_colors             = 1;
 	e_tree->priv->horizontal_draw_grid               = 1;
 	e_tree->priv->vertical_draw_grid                 = 1;
 	e_tree->priv->draw_focus                         = 1;
 	e_tree->priv->cursor_mode                        = E_CURSOR_SIMPLE;
 	e_tree->priv->length_threshold                   = 200;
+	e_tree->priv->uniform_row_height                 = FALSE;
 
 	e_tree->priv->row_selection_active               = FALSE;
 	e_tree->priv->horizontal_scrolling               = FALSE;
@@ -281,14 +327,22 @@ e_tree_init (GtkObject *object)
 	e_tree->priv->drop_row                           = -1;
 	e_tree->priv->drop_path                          = NULL;
 	e_tree->priv->drop_col                           = -1;
+	e_tree->priv->drop_highlight                     = NULL;
+
+	e_tree->priv->last_drop_x                        = 0;
+	e_tree->priv->last_drop_y                        = 0;
+	e_tree->priv->last_drop_time                     = 0;
+	e_tree->priv->last_drop_context                  = NULL;
+
+	e_tree->priv->hover_x                            = 0;
+	e_tree->priv->hover_y                            = 0;
 
 	e_tree->priv->drag_row                           = -1;
 	e_tree->priv->drag_path                          = NULL;
 	e_tree->priv->drag_col                           = -1;
 
 	e_tree->priv->site                               = NULL;
-	e_tree->priv->drag_source_button_press_event_id  = 0;
-	e_tree->priv->drag_source_motion_notify_event_id = 0;
+	e_tree->priv->do_drag                            = FALSE;
 
 #ifdef E_TREE_USE_TREE_SELECTION
 	e_tree->priv->selection                          = E_SELECTION_MODEL(e_tree_selection_model_new());
@@ -389,7 +443,7 @@ e_tree_setup_header (ETree *e_tree)
 		"full_header", e_tree->priv->full_header,
 		"sort_info", e_tree->priv->sort_info,
 		"dnd_code", pointer,
-		/*		"table", e_tree, FIXME*/
+		"tree", e_tree,
 		NULL);
 
 	g_free(pointer);
@@ -483,10 +537,12 @@ static void
 item_cursor_activated (ETableItem *eti, int row, ETree *et)
 {
 	ETreePath path = e_tree_table_adapter_node_at_row(et->priv->etta, row);
-	path = e_tree_sorted_view_to_model_path(et->priv->sorted, path);
+	if (path)
+		path = e_tree_sorted_view_to_model_path(et->priv->sorted, path);
 	gtk_signal_emit (GTK_OBJECT (et),
 			 et_signals [CURSOR_ACTIVATED],
 			 row, path);
+	d(g_print("%s: Emitted CURSOR_ACTIVATED signal on row: %d and path: 0x%p\n", __FUNCTION__, row, path));
 }
 
 static void
@@ -534,27 +590,40 @@ item_key_press (ETableItem *eti, int row, int col, GdkEvent *event, ETree *et)
 
 	switch (key->keyval) {
 	case GDK_Page_Down:
+	case GDK_KP_Page_Down:
 		vadj = gtk_layout_get_vadjustment (GTK_LAYOUT (et->priv->table_canvas));
 		y = CLAMP(vadj->value + (2 * vadj->page_size - 50), 0, vadj->upper);
 		y -= vadj->value;
 		e_tree_get_cell_at (et, 30, y, &row_local, &col_local);
+
+		if (row_local == -1)
+			row_local = e_table_model_row_count (E_TABLE_MODEL(et->priv->etta)) - 1;
+
 		row_local = e_tree_view_to_model_row (et, row_local);
 		col_local = e_selection_model_cursor_col (E_SELECTION_MODEL (et->priv->selection));
 		e_selection_model_select_as_key_press (E_SELECTION_MODEL (et->priv->selection), row_local, col_local, key->state);
+
 		return_val = 1;
 		break;
 	case GDK_Page_Up:
+	case GDK_KP_Page_Up:
 		vadj = gtk_layout_get_vadjustment (GTK_LAYOUT (et->priv->table_canvas));
 		y = CLAMP(vadj->value - (vadj->page_size - 50), 0, vadj->upper);
 		y -= vadj->value;
 		e_tree_get_cell_at (et, 30, y, &row_local, &col_local);
+
+		if (row_local == -1)
+			row_local = e_table_model_row_count (E_TABLE_MODEL(et->priv->etta)) - 1;
+
 		row_local = e_tree_view_to_model_row (et, row_local);
 		col_local = e_selection_model_cursor_col (E_SELECTION_MODEL (et->priv->selection));
 		e_selection_model_select_as_key_press (E_SELECTION_MODEL (et->priv->selection), row_local, col_local, key->state);
+
 		return_val = 1;
 		break;
 	case '=':
 	case GDK_Right:
+	case GDK_KP_Right:
 		if (row != -1) {
 			path = e_tree_table_adapter_node_at_row(et->priv->etta, row);
 			if (path)
@@ -564,6 +633,7 @@ item_key_press (ETableItem *eti, int row, int col, GdkEvent *event, ETree *et)
 		break;
 	case '-':
 	case GDK_Left:
+	case GDK_KP_Left:
 		if (row != -1) {
 			path = e_tree_table_adapter_node_at_row(et->priv->etta, row);
 			if (path)
@@ -582,6 +652,22 @@ item_key_press (ETableItem *eti, int row, int col, GdkEvent *event, ETree *et)
 	return return_val;
 }
 
+static gint
+item_start_drag (ETableItem *eti, int row, int col, GdkEvent *event, ETree *et)
+{
+	ETreePath path;
+	gint return_val = 0;
+
+	path = e_tree_table_adapter_node_at_row(et->priv->etta, row);
+	path = e_tree_sorted_view_to_model_path(et->priv->sorted, path);
+
+	gtk_signal_emit (GTK_OBJECT (et),
+			 et_signals [START_DRAG],
+			 row, path, col, event, &return_val);
+
+	return return_val;
+}
+
 static void
 et_selection_model_selection_change (ETableSelectionModel *etsm, ETable *et)
 {
@@ -603,6 +689,7 @@ et_build_item (ETree *et)
 					 "drawfocus", et->priv->draw_focus,
 					 "cursor_mode", et->priv->cursor_mode,
 					 "length_threshold", et->priv->length_threshold,
+					 "uniform_row_height", et->priv->uniform_row_height,
 					 NULL);
 
 	gtk_signal_connect (GTK_OBJECT (et->priv->item), "cursor_change",
@@ -617,6 +704,8 @@ et_build_item (ETree *et)
 			    GTK_SIGNAL_FUNC (item_click), et);
 	gtk_signal_connect (GTK_OBJECT (et->priv->item), "key_press",
 			    GTK_SIGNAL_FUNC (item_key_press), et);
+	gtk_signal_connect (GTK_OBJECT (et->priv->item), "start_drag",
+			    GTK_SIGNAL_FUNC (item_start_drag), et);
 }
 
 static void
@@ -628,16 +717,54 @@ et_canvas_realize (GtkWidget *canvas, ETree *e_tree)
 		NULL);
 }
 
-static void
-et_canvas_button_press (GtkWidget *canvas, GdkEvent *event, ETree *e_tree)
+static gint
+et_canvas_root_event (GnomeCanvasItem *root, GdkEvent *event, ETree *e_tree)
 {
-	if (GTK_WIDGET_HAS_FOCUS(canvas)) {
-		GnomeCanvasItem *item = GNOME_CANVAS(canvas)->focused_item;
+	switch (event->type) {
+	case GDK_BUTTON_PRESS:
+	case GDK_2BUTTON_PRESS:
+	case GDK_BUTTON_RELEASE:
+		if (event->button.button != 4 && event->button.button != 5) {
+			if (GTK_WIDGET_HAS_FOCUS(root->canvas)) {
+				GnomeCanvasItem *item = GNOME_CANVAS(root->canvas)->focused_item;
 
-		if (E_IS_TABLE_ITEM(item)) {
-			e_table_item_leave_edit(E_TABLE_ITEM(item));
+				if (E_IS_TABLE_ITEM(item)) {
+					e_table_item_leave_edit_(E_TABLE_ITEM(item));
+					return TRUE;
+				}
+			}
 		}
+		break;
+	default:
+		break;
 	}
+
+	return FALSE;
+}
+
+/* Handler for focus events in the table_canvas; we have to repaint ourselves
+ * and give the focus to some ETableItem.
+ */
+static gint
+table_canvas_focus_event_cb (GtkWidget *widget, GdkEventFocus *event, gpointer data)
+{
+	GnomeCanvas *canvas;
+	ETree *tree;
+
+	gtk_widget_queue_draw (widget);
+
+	if (!event->in)
+		return TRUE;
+
+	canvas = GNOME_CANVAS (widget);
+	tree = E_TREE (data);
+
+	if (!canvas->focused_item) {
+		e_table_item_set_cursor (E_TABLE_ITEM (tree->priv->item), 0, 0);
+		gnome_canvas_item_grab_focus (tree->priv->item);
+	}
+
+	return TRUE;
 }
 
 static void
@@ -649,22 +776,22 @@ e_tree_setup_table (ETree *e_tree)
 		GTK_SIGNAL_FUNC (tree_canvas_size_allocate), e_tree);
 	gtk_signal_connect (
 		GTK_OBJECT (e_tree->priv->table_canvas), "focus_in_event",
-		GTK_SIGNAL_FUNC (gtk_widget_queue_draw), e_tree);
+		GTK_SIGNAL_FUNC (table_canvas_focus_event_cb), e_tree);
 	gtk_signal_connect (
 		GTK_OBJECT (e_tree->priv->table_canvas), "focus_out_event",
-		GTK_SIGNAL_FUNC (gtk_widget_queue_draw), e_tree);
+		GTK_SIGNAL_FUNC (table_canvas_focus_event_cb), e_tree);
 
 	gtk_signal_connect (
-		GTK_OBJECT (e_tree), "drag_begin",
+		GTK_OBJECT (e_tree->priv->table_canvas), "drag_begin",
 		GTK_SIGNAL_FUNC (et_drag_begin), e_tree);
 	gtk_signal_connect (
-		GTK_OBJECT (e_tree), "drag_end",
+		GTK_OBJECT (e_tree->priv->table_canvas), "drag_end",
 		GTK_SIGNAL_FUNC (et_drag_end), e_tree);
 	gtk_signal_connect (
-		GTK_OBJECT (e_tree), "drag_data_get",
+		GTK_OBJECT (e_tree->priv->table_canvas), "drag_data_get",
 		GTK_SIGNAL_FUNC (et_drag_data_get), e_tree);
 	gtk_signal_connect (
-		GTK_OBJECT (e_tree), "drag_data_delete",
+		GTK_OBJECT (e_tree->priv->table_canvas), "drag_data_delete",
 		GTK_SIGNAL_FUNC (et_drag_data_delete), e_tree);
 	gtk_signal_connect (
 		GTK_OBJECT (e_tree), "drag_motion",
@@ -698,8 +825,8 @@ e_tree_setup_table (ETree *e_tree)
 		GTK_OBJECT(e_tree->priv->table_canvas), "realize",
 		GTK_SIGNAL_FUNC(et_canvas_realize), e_tree);
 	gtk_signal_connect (
-		GTK_OBJECT(e_tree->priv->table_canvas), "button_press_event",
-		GTK_SIGNAL_FUNC(et_canvas_button_press), e_tree);
+		GTK_OBJECT(gnome_canvas_root (e_tree->priv->table_canvas)), "event",
+		GTK_SIGNAL_FUNC(et_canvas_root_event), e_tree);
 
 	et_build_item(e_tree);
 }
@@ -718,9 +845,10 @@ e_tree_set_state_object(ETree *e_tree, ETableState *state)
 	if (e_tree->priv->sort_info)
 		gtk_object_unref(GTK_OBJECT(e_tree->priv->sort_info));
 
-	if (state->sort_info)
+	if (state->sort_info) {
 		e_tree->priv->sort_info = e_table_sort_info_duplicate(state->sort_info);
-	else
+		e_table_sort_info_set_can_group (e_tree->priv->sort_info, FALSE);
+	} else
 		e_tree->priv->sort_info = NULL;
 
 	if (e_tree->priv->header_item)
@@ -810,7 +938,8 @@ e_tree_get_state_object (ETree *e_tree)
 
 	state = e_table_state_new();
 	state->sort_info = e_tree->priv->sort_info;
-	gtk_object_ref(GTK_OBJECT(state->sort_info));
+	if (state->sort_info)
+		gtk_object_ref(GTK_OBJECT(state->sort_info));
 
 	state->col_count = e_table_header_count (e_tree->priv->header);
 	full_col_count = e_table_header_count (e_tree->priv->full_header);
@@ -861,6 +990,12 @@ e_tree_save_state (ETree      *e_tree,
 	gtk_object_unref(GTK_OBJECT(state));
 }
 
+ETableSpecification *
+e_tree_get_spec (ETree *e_tree)
+{
+	return e_tree->priv->spec;
+}
+
 static void
 et_table_model_changed (ETableModel *model, ETree *et)
 {
@@ -936,6 +1071,8 @@ et_real_construct (ETree *e_tree, ETreeModel *etm, ETableExtras *ete,
 	e_tree->priv->sort_info = state->sort_info;
 	gtk_object_ref (GTK_OBJECT (e_tree->priv->sort_info));
 
+	e_table_sort_info_set_can_group (e_tree->priv->sort_info, FALSE);
+
 	gtk_object_set(GTK_OBJECT(e_tree->priv->header),
 		       "sort_info", e_tree->priv->sort_info,
 		       NULL);
@@ -977,6 +1114,8 @@ et_real_construct (ETree *e_tree, ETreeModel *etm, ETableExtras *ete,
 
 	gtk_layout_get_vadjustment (GTK_LAYOUT (e_tree->priv->table_canvas))->step_increment = 20;
 	gtk_adjustment_changed(gtk_layout_get_vadjustment (GTK_LAYOUT (e_tree->priv->table_canvas)));
+	gtk_layout_get_hadjustment (GTK_LAYOUT (e_tree->priv->table_canvas))->step_increment = 20;
+	gtk_adjustment_changed(gtk_layout_get_hadjustment (GTK_LAYOUT (e_tree->priv->table_canvas)));
 
 	if (!specification->no_headers) {
 		/*
@@ -1219,6 +1358,9 @@ et_get_arg (GtkObject *o, GtkArg *arg, guint arg_id)
 			GTK_VALUE_OBJECT (*arg) = GTK_OBJECT (etree->priv->etta);
 		}
 		break;
+	case ARG_UNIFORM_ROW_HEIGHT:
+		GTK_VALUE_BOOL (*arg) = etree->priv->uniform_row_height;
+		break;
 
 	default:
 		break;
@@ -1271,6 +1413,15 @@ et_set_arg (GtkObject *o, GtkArg *arg, guint arg_id)
 					       NULL);
 		}
 		break;
+
+	case ARG_UNIFORM_ROW_HEIGHT:
+		etree->priv->uniform_row_height = GTK_VALUE_BOOL (*arg);
+		if (etree->priv->item) {
+			gnome_canvas_item_set (GNOME_CANVAS_ITEM(etree->priv->item),
+					       "uniform_row_height", GTK_VALUE_BOOL (*arg),
+					       NULL);
+		}
+		break;
 	}
 }
 
@@ -1283,6 +1434,10 @@ set_scroll_adjustments   (ETree *tree,
 		vadjustment->step_increment = 20;
 		gtk_adjustment_changed(vadjustment);
 	}
+	if (hadjustment != NULL) {
+		hadjustment->step_increment = 20;
+		gtk_adjustment_changed(hadjustment);
+	}
 
 	gtk_layout_set_hadjustment (GTK_LAYOUT(tree->priv->table_canvas),
 				    hadjustment);
@@ -1465,6 +1620,12 @@ e_tree_get_tooltip (ETree *et)
 	return E_CANVAS(et->priv->table_canvas)->tooltip_window;
 }
 
+void
+e_tree_right_click_up (ETree *et)
+{
+	e_selection_model_right_click_up(et->priv->selection);
+}
+
 /**
  * e_tree_get_model:
  * @et: the ETree
@@ -1569,6 +1730,7 @@ struct _GtkDragSourceInfo
 
 /* Drag & drop stuff. */
 /* Target */
+
 void
 e_tree_drag_get_data (ETree         *tree,
 		      int             row,
@@ -1591,6 +1753,29 @@ e_tree_drag_get_data (ETree         *tree,
 
 }
 
+#if 0
+static void
+e_tree_request_hightlight_redraw (ETree *tree)
+{
+	int row = tree->drop_highlight_row;
+	int col = tree->drop_highlight_col;
+
+	if (row != -1) {
+		int x, y, width, height;
+		if (col == -1) {
+			e_tree_get_cell_geometry (tree, row, 0, &x, &y, &width, &height);
+			x = 0;
+			width = tree->allocation.width;
+		} else {
+			e_tree_get_cell_geometry (tree, row, col, &x, &y, &width, &height);
+			x += GTK_LAYOUT(tree->priv->table_canvas)->hadjustment->value;
+		}
+		y += GTK_LAYOUT(tree->priv->table_canvas)->vadjustment->value;
+		gnome_canvas_request_redraw (tree->priv->table_canvas, x, y, x + width - 1, y + height - 1);
+	}
+}
+#endif
+
 /**
  * e_tree_drag_highlight:
  * @tree:
@@ -1598,14 +1783,48 @@ e_tree_drag_get_data (ETree         *tree,
  * @col:
  *
  * Set col to -1 to highlight the entire row.
+ * Set row to -1 to turn off the highlight.
  */
 void
 e_tree_drag_highlight (ETree *tree,
-			int     row,
-			int     col)
+		       int     row,
+		       int     col)
 {
 	g_return_if_fail(tree != NULL);
 	g_return_if_fail(E_IS_TREE(tree));
+
+	if (row != -1) {
+		int x, y, width, height;
+		if (col == -1) {
+			e_tree_get_cell_geometry (tree, row, 0, &x, &y, &width, &height);
+			x = 0;
+			width = GTK_WIDGET (tree->priv->table_canvas)->allocation.width;
+		} else {
+			e_tree_get_cell_geometry (tree, row, col, &x, &y, &width, &height);
+			x += GTK_LAYOUT(tree->priv->table_canvas)->hadjustment->value;
+		}
+		y += GTK_LAYOUT(tree->priv->table_canvas)->vadjustment->value;
+
+		if (tree->priv->drop_highlight == NULL) {
+			tree->priv->drop_highlight =
+				gnome_canvas_item_new (gnome_canvas_root (tree->priv->table_canvas),
+						       gnome_canvas_rect_get_type (),
+						       "fill_color", NULL,
+						       /*						       "outline_color", "black",
+						       "width_pixels", 1,*/
+						       "outline_color_gdk", &(GTK_WIDGET (tree)->style->fg[GTK_STATE_NORMAL]),
+						       NULL);
+		}
+		gnome_canvas_item_set (tree->priv->drop_highlight,
+				       "x1", (double) x,
+				       "x2", (double) x + width - 1,
+				       "y1", (double) y,
+				       "y2", (double) y + height - 1,
+				       NULL);
+	} else {
+		gtk_object_destroy (GTK_OBJECT (tree->priv->drop_highlight));
+		tree->priv->drop_highlight = NULL;
+	}
 }
 
 void
@@ -1613,6 +1832,11 @@ e_tree_drag_unhighlight (ETree *tree)
 {
 	g_return_if_fail(tree != NULL);
 	g_return_if_fail(E_IS_TREE(tree));
+
+	if (tree->priv->drop_highlight) {
+		gtk_object_destroy (GTK_OBJECT (tree->priv->drop_highlight));
+		tree->priv->drop_highlight = NULL;
+	}
 }
 
 void e_tree_drag_dest_set   (ETree               *tree,
@@ -1661,6 +1885,40 @@ e_tree_drag_dest_unset (GtkWidget *widget)
 
 /* Source side */
 
+static gint
+et_real_start_drag (ETree *tree, int row, ETreePath path, int col, GdkEvent *event)
+{
+	GtkDragSourceInfo *info;
+	GdkDragContext *context;
+	ETreeDragSourceSite *site;
+
+	if (tree->priv->do_drag) {
+		site = tree->priv->site;
+
+		site->state = 0;
+		context = e_tree_drag_begin (tree, row, col,
+					     site->target_list,
+					     site->actions,
+					     1, event);
+
+		if (context) {
+			info = g_dataset_get_data (context, "gtk-info");
+
+			if (info && !info->icon_window) {
+				if (site->pixmap)
+					gtk_drag_set_icon_pixmap (context,
+								  site->colormap,
+								  site->pixmap,
+								  site->mask, -2, -2);
+				else
+					gtk_drag_set_icon_default (context);
+			}
+		}
+		return TRUE;
+	}
+	return FALSE;
+}
+
 void
 e_tree_drag_source_set  (ETree               *tree,
 			 GdkModifierType       start_button_mask,
@@ -1677,6 +1935,8 @@ e_tree_drag_source_set  (ETree               *tree,
 	canvas = GTK_WIDGET(tree->priv->table_canvas);
 	site = tree->priv->site;
 
+	tree->priv->do_drag = TRUE;
+
 	gtk_widget_add_events (canvas,
 			       gtk_widget_get_events (canvas) |
 			       GDK_BUTTON_PRESS_MASK | GDK_BUTTON_RELEASE_MASK |
@@ -1687,16 +1947,6 @@ e_tree_drag_source_set  (ETree               *tree,
 			gtk_target_list_unref (site->target_list);
 	} else {
 		site = g_new0 (ETreeDragSourceSite, 1);
-
-		tree->priv->drag_source_button_press_event_id =
-			gtk_signal_connect (GTK_OBJECT (canvas), "button_press_event",
-					    GTK_SIGNAL_FUNC (e_tree_drag_source_event_cb),
-					    tree);
-		tree->priv->drag_source_motion_notify_event_id =
-			gtk_signal_connect (GTK_OBJECT (canvas), "motion_notify_event",
-					    GTK_SIGNAL_FUNC (e_tree_drag_source_event_cb),
-					    tree);
-
 		tree->priv->site = site;
 	}
 
@@ -1721,12 +1971,6 @@ e_tree_drag_source_unset (ETree *tree)
 	site = tree->priv->site;
 
 	if (site) {
-		gtk_signal_disconnect (
-			GTK_OBJECT (tree->priv->table_canvas),
-			tree->priv->drag_source_button_press_event_id);
-		gtk_signal_disconnect (
-			GTK_OBJECT (tree->priv->table_canvas),
-			tree->priv->drag_source_motion_notify_event_id);
 		g_free (site);
 		tree->priv->site = NULL;
 	}
@@ -1756,7 +2000,7 @@ e_tree_drag_begin (ETree            *tree,
 	tree->priv->drag_path = path;
 	tree->priv->drag_col = col;
 
-	return gtk_drag_begin(GTK_WIDGET(tree),
+	return gtk_drag_begin(GTK_WIDGET (tree->priv->table_canvas),
 			      targets,
 			      actions,
 			      button,
@@ -1787,6 +2031,11 @@ e_tree_get_cell_at (ETree *tree,
 	/* FIXME it would be nice if it could handle a NULL row_return or
 	 * col_return gracefully.  */
 
+	if (row_return)
+		*row_return = -1;
+	if (col_return)
+		*col_return = -1;
+
 	x += GTK_LAYOUT(tree->priv->table_canvas)->hadjustment->value;
 	y += GTK_LAYOUT(tree->priv->table_canvas)->vadjustment->value;
 	e_table_item_compute_location(E_TABLE_ITEM(tree->priv->item), &x, &y, row_return, col_return);
@@ -1884,40 +2133,28 @@ et_drag_data_delete(GtkWidget *widget,
 			 context);
 }
 
-static void
-et_drag_leave(GtkWidget *widget,
-	      GdkDragContext *context,
-	      guint time,
-	      ETree *et)
-{
-	gtk_signal_emit (GTK_OBJECT (et),
-			 et_signals [TREE_DRAG_LEAVE],
-			 et->priv->drop_row,
-			 et->priv->drop_path,
-			 et->priv->drop_col,
-			 context,
-			 time);
-	et->priv->drop_row = -1;
-	et->priv->drop_col = -1;
-}
-
 static gboolean
-et_drag_motion(GtkWidget *widget,
+do_drag_motion(ETree *et,
 	       GdkDragContext *context,
 	       gint x,
 	       gint y,
-	       guint time,
-	       ETree *et)
+	       guint time)
 {
-	gboolean ret_val;
+	gboolean ret_val = FALSE;
 	int row, col;
 	ETreePath path;
+	GtkWidget *widget;
+
+	widget = GTK_WIDGET (et);
+
+	x -= widget->allocation.x;
+	y -= widget->allocation.y;
 	e_tree_get_cell_at (et,
 			    x,
 			    y,
 			    &row,
 			    &col);
-	if (row != et->priv->drop_row && col != et->priv->drop_row) {
+	if (row != et->priv->drop_row && col != et->priv->drop_col) {
 		gtk_signal_emit (GTK_OBJECT (et),
 				 et_signals [TREE_DRAG_LEAVE],
 				 et->priv->drop_row,
@@ -1943,6 +2180,190 @@ et_drag_motion(GtkWidget *widget,
 			 y,
 			 time,
 			 &ret_val);
+
+	return ret_val;
+}
+
+static gboolean
+scroll_timeout (gpointer data)
+{
+	ETree *et = data;
+	int dy;
+	GtkAdjustment *v;
+	double value;
+
+	if (et->priv->scroll_down)
+		dy = 20;
+	else
+		dy = -20;
+
+	v = GTK_LAYOUT(et->priv->table_canvas)->vadjustment;
+
+	value = v->value;
+
+	gtk_adjustment_set_value(v, CLAMP(v->value + dy, v->lower, v->upper - v->page_size));
+
+	if (v->value != value)
+		do_drag_motion(et,
+			       et->priv->last_drop_context,
+			       et->priv->last_drop_x,
+			       et->priv->last_drop_y,
+			       et->priv->last_drop_time);
+			       
+
+	return TRUE;
+}
+
+static void
+scroll_on (ETree *et, gboolean down)
+{
+	if (et->priv->scroll_idle_id == 0 || down != et->priv->scroll_down) {
+		if (et->priv->scroll_idle_id != 0)
+			g_source_remove (et->priv->scroll_idle_id);
+		et->priv->scroll_down = down;
+		et->priv->scroll_idle_id = g_timeout_add (100, scroll_timeout, et);
+	}
+}
+
+static void
+scroll_off (ETree *et)
+{
+	if (et->priv->scroll_idle_id) {
+		g_source_remove (et->priv->scroll_idle_id);
+		et->priv->scroll_idle_id = 0;
+	}
+}
+ 
+static gboolean
+hover_timeout (gpointer data)
+{
+	ETree *et = data;
+	GtkWidget *widget = data;
+	int x = et->priv->hover_x;
+	int y = et->priv->hover_y;
+	int row, col;
+	ETreePath path;
+
+	x -= widget->allocation.x;
+	y -= widget->allocation.y;
+	e_tree_get_cell_at (et,
+			    x,
+			    y,
+			    &row,
+			    &col);
+
+	path = e_tree_table_adapter_node_at_row(et->priv->etta, row);
+	if (path && e_tree_model_node_is_expandable (E_TREE_MODEL (et->priv->sorted), path)) {
+		e_tree_table_adapter_node_set_expanded (et->priv->etta, path, TRUE);
+	}
+
+	return TRUE;
+}
+
+static void
+hover_on (ETree *et, int x, int y)
+{
+	et->priv->hover_x = x;
+	et->priv->hover_y = y;
+	if (et->priv->hover_idle_id != 0)
+		g_source_remove (et->priv->hover_idle_id);
+	et->priv->hover_idle_id = g_timeout_add (500, hover_timeout, et);
+}
+
+static void
+hover_off (ETree *et)
+{
+	if (et->priv->hover_idle_id) {
+		g_source_remove (et->priv->hover_idle_id);
+		et->priv->hover_idle_id = 0;
+	}
+}
+
+static void
+context_destroyed (gpointer data)
+{
+	ETree *et = data;
+	et->priv->last_drop_x       = 0;
+	et->priv->last_drop_y       = 0;
+	et->priv->last_drop_time    = 0;
+	et->priv->last_drop_context = NULL;
+	scroll_off (et);
+	hover_off (et);
+	gtk_object_unref (GTK_OBJECT (et));
+}
+
+static void
+context_connect (ETree *et, GdkDragContext *context)
+{
+	if (g_dataset_get_data (context, "e-tree") == NULL) {
+		gtk_object_ref (GTK_OBJECT (et));
+		g_dataset_set_data_full (context, "e-tree", et, context_destroyed);
+	}
+}
+
+static void
+et_drag_leave(GtkWidget *widget,
+	      GdkDragContext *context,
+	      guint time,
+	      ETree *et)
+{
+	gtk_signal_emit (GTK_OBJECT (et),
+			 et_signals [TREE_DRAG_LEAVE],
+			 et->priv->drop_row,
+			 et->priv->drop_path,
+			 et->priv->drop_col,
+			 context,
+			 time);
+	et->priv->drop_row = -1;
+	et->priv->drop_col = -1;
+
+	scroll_off (et);
+	hover_off (et);
+}
+
+static gboolean
+et_drag_motion(GtkWidget *widget,
+	       GdkDragContext *context,
+	       gint x,
+	       gint y,
+	       guint time,
+	       ETree *et)
+{
+	int ret_val;
+
+	et->priv->last_drop_x = x;
+	et->priv->last_drop_y = y;
+	et->priv->last_drop_time = time;
+	et->priv->last_drop_context = context;
+	context_connect (et, context);
+
+	if (et->priv->hover_idle_id != 0) {
+		if (abs (et->priv->hover_x - x) > 3 ||
+		    abs (et->priv->hover_y - y) > 3) {
+			hover_on (et, x, y);
+		}
+	} else {
+		hover_on (et, x, y);
+	}
+
+	ret_val = do_drag_motion (et,
+				  context,
+				  x,
+				  y,
+				  time);
+
+	x -= widget->allocation.x;
+	y -= widget->allocation.y;
+
+	if (y < 20 || y > widget->allocation.height - 20) {
+		if (y < 20)
+			scroll_on (et, FALSE);
+		else
+			scroll_on (et, TRUE);
+	} else {
+		scroll_off (et);
+	}
+
 	return ret_val;
 }
 
@@ -1954,9 +2375,11 @@ et_drag_drop(GtkWidget *widget,
 	     guint time,
 	     ETree *et)
 {
-	gboolean ret_val;
+	gboolean ret_val = FALSE;
 	int row, col;
 	ETreePath path;
+	y -= widget->allocation.y;
+	x -= widget->allocation.x;
 	e_tree_get_cell_at(et,
 			   x,
 			   y,
@@ -2000,6 +2423,8 @@ et_drag_drop(GtkWidget *widget,
 	et->priv->drop_row = -1;
 	et->priv->drop_path = NULL;
 	et->priv->drop_col = -1;
+
+	scroll_off (et);
 	return ret_val;
 }
 
@@ -2015,6 +2440,8 @@ et_drag_data_received(GtkWidget *widget,
 {
 	int row, col;
 	ETreePath path;
+	y -= widget->allocation.y;
+	x -= widget->allocation.x;
 	e_tree_get_cell_at(et,
 			   x,
 			   y,
@@ -2035,82 +2462,6 @@ et_drag_data_received(GtkWidget *widget,
 			 time);
 }
 
-static gint
-e_tree_drag_source_event_cb (GtkWidget      *widget,
-			      GdkEvent       *event,
-			      ETree         *tree)
-{
-	ETreeDragSourceSite *site;
-	site = tree->priv->site;
-
-	switch (event->type) {
-	case GDK_BUTTON_PRESS:
-		if ((GDK_BUTTON1_MASK << (event->button.button - 1)) & site->start_button_mask) {
-			int row, col;
-			e_tree_get_cell_at(tree, event->button.x, event->button.y, &row, &col);
-			if (row >= 0 && col >= 0) {
-				site->state |= (GDK_BUTTON1_MASK << (event->button.button - 1));
-				site->x = event->button.x;
-				site->y = event->button.y;
-				site->row = row;
-				site->col = col;
-			}
-		}
-		break;
-
-	case GDK_BUTTON_RELEASE:
-		if ((GDK_BUTTON1_MASK << (event->button.button - 1)) & site->start_button_mask) {
-			site->state &= ~(GDK_BUTTON1_MASK << (event->button.button - 1));
-		}
-		break;
-
-	case GDK_MOTION_NOTIFY:
-		if (site->state & event->motion.state & site->start_button_mask) {
-			/* FIXME: This is really broken and can leave us
-			 * with a stuck grab
-			 */
-			int i;
-			for (i=1; i<6; i++) {
-				if (site->state & event->motion.state &
-				    GDK_BUTTON1_MASK << (i - 1))
-					break;
-			}
-
-			if (MAX (abs (site->x - event->motion.x),
-				 abs (site->y - event->motion.y)) > 3) {
-				GtkDragSourceInfo *info;
-				GdkDragContext *context;
-
-				site->state = 0;
-				context = e_tree_drag_begin (tree, site->row, site->col,
-							     site->target_list,
-							     site->actions,
-							     i, event);
-
-
-				info = g_dataset_get_data (context, "gtk-info");
-
-				if (!info->icon_window) {
-					if (site->pixmap)
-						gtk_drag_set_icon_pixmap (context,
-									  site->colormap,
-									  site->pixmap,
-									  site->mask, -2, -2);
-					else
-						gtk_drag_set_icon_default (context);
-				}
-
-				return TRUE;
-			}
-		}
-		break;
-
-	default:			/* hit for 2/3BUTTON_PRESS */
-		break;
-	}
-	return FALSE;
-}
-
 static void
 e_tree_class_init (ETreeClass *class)
 {
@@ -2118,27 +2469,28 @@ e_tree_class_init (ETreeClass *class)
 	GtkWidgetClass *widget_class;
 	GtkContainerClass *container_class;
 
-	object_class = (GtkObjectClass *) class;
-	widget_class = (GtkWidgetClass *) class;
-	container_class = (GtkContainerClass *) class;
+	object_class                   = (GtkObjectClass *) class;
+	widget_class                   = (GtkWidgetClass *) class;
+	container_class                = (GtkContainerClass *) class;
 
-	parent_class = gtk_type_class (PARENT_TYPE);
+	parent_class                   = gtk_type_class (PARENT_TYPE);
 
-	object_class->destroy           = et_destroy;
-	object_class->set_arg           = et_set_arg;
-	object_class->get_arg           = et_get_arg;
+	object_class->destroy          = et_destroy;
+	object_class->set_arg          = et_set_arg;
+	object_class->get_arg          = et_get_arg;
 
-	widget_class->grab_focus = et_grab_focus;
+	widget_class->grab_focus       = et_grab_focus;
 
-	container_class->focus = et_focus;
+	container_class->focus         = et_focus;
 
-	class->cursor_change            = NULL;
-	class->cursor_activated            = NULL;
-	class->selection_change         = NULL;
-	class->double_click             = NULL;
-	class->right_click              = NULL;
-	class->click                    = NULL;
-	class->key_press                = NULL;
+	class->cursor_change           = NULL;
+	class->cursor_activated        = NULL;
+	class->selection_change        = NULL;
+	class->double_click            = NULL;
+	class->right_click             = NULL;
+	class->click                   = NULL;
+	class->key_press               = NULL;
+	class->start_drag              = et_real_start_drag;
 
 	class->tree_drag_begin         = NULL;
 	class->tree_drag_end           = NULL;
@@ -2153,7 +2505,7 @@ e_tree_class_init (ETreeClass *class)
 	et_signals [CURSOR_CHANGE] =
 		gtk_signal_new ("cursor_change",
 				GTK_RUN_LAST,
-				object_class->type,
+				E_OBJECT_CLASS_TYPE (object_class),
 				GTK_SIGNAL_OFFSET (ETreeClass, cursor_change),
 				gtk_marshal_NONE__INT_POINTER,
 				GTK_TYPE_NONE, 2, GTK_TYPE_INT, GTK_TYPE_POINTER);
@@ -2161,7 +2513,7 @@ e_tree_class_init (ETreeClass *class)
 	et_signals [CURSOR_ACTIVATED] =
 		gtk_signal_new ("cursor_activated",
 				GTK_RUN_LAST,
-				object_class->type,
+				E_OBJECT_CLASS_TYPE (object_class),
 				GTK_SIGNAL_OFFSET (ETreeClass, cursor_activated),
 				gtk_marshal_NONE__INT_POINTER,
 				GTK_TYPE_NONE, 2, GTK_TYPE_INT, GTK_TYPE_POINTER);
@@ -2169,7 +2521,7 @@ e_tree_class_init (ETreeClass *class)
 	et_signals [SELECTION_CHANGE] =
 		gtk_signal_new ("selection_change",
 				GTK_RUN_LAST,
-				object_class->type,
+				E_OBJECT_CLASS_TYPE (object_class),
 				GTK_SIGNAL_OFFSET (ETreeClass, selection_change),
 				gtk_marshal_NONE__NONE,
 				GTK_TYPE_NONE, 0);
@@ -2177,7 +2529,7 @@ e_tree_class_init (ETreeClass *class)
 	et_signals [DOUBLE_CLICK] =
 		gtk_signal_new ("double_click",
 				GTK_RUN_LAST,
-				object_class->type,
+				E_OBJECT_CLASS_TYPE (object_class),
 				GTK_SIGNAL_OFFSET (ETreeClass, double_click),
 				e_marshal_NONE__INT_POINTER_INT_POINTER,
 				GTK_TYPE_NONE, 4, GTK_TYPE_INT, GTK_TYPE_POINTER, GTK_TYPE_INT, GTK_TYPE_GDK_EVENT);
@@ -2185,7 +2537,7 @@ e_tree_class_init (ETreeClass *class)
 	et_signals [RIGHT_CLICK] =
 		gtk_signal_new ("right_click",
 				GTK_RUN_LAST,
-				object_class->type,
+				E_OBJECT_CLASS_TYPE (object_class),
 				GTK_SIGNAL_OFFSET (ETreeClass, right_click),
 				e_marshal_INT__INT_POINTER_INT_POINTER,
 				GTK_TYPE_INT, 4, GTK_TYPE_INT, GTK_TYPE_POINTER, GTK_TYPE_INT, GTK_TYPE_GDK_EVENT);
@@ -2193,7 +2545,7 @@ e_tree_class_init (ETreeClass *class)
 	et_signals [CLICK] =
 		gtk_signal_new ("click",
 				GTK_RUN_LAST,
-				object_class->type,
+				E_OBJECT_CLASS_TYPE (object_class),
 				GTK_SIGNAL_OFFSET (ETreeClass, click),
 				e_marshal_INT__INT_POINTER_INT_POINTER,
 				GTK_TYPE_INT, 4, GTK_TYPE_INT, GTK_TYPE_POINTER, GTK_TYPE_INT, GTK_TYPE_GDK_EVENT);
@@ -2201,15 +2553,23 @@ e_tree_class_init (ETreeClass *class)
 	et_signals [KEY_PRESS] =
 		gtk_signal_new ("key_press",
 				GTK_RUN_LAST,
-				object_class->type,
+				E_OBJECT_CLASS_TYPE (object_class),
 				GTK_SIGNAL_OFFSET (ETreeClass, key_press),
 				e_marshal_INT__INT_POINTER_INT_POINTER,
 				GTK_TYPE_INT, 4, GTK_TYPE_INT, GTK_TYPE_POINTER, GTK_TYPE_INT, GTK_TYPE_GDK_EVENT);
 
+	et_signals [START_DRAG] =
+		gtk_signal_new ("start_drag",
+				GTK_RUN_LAST,
+				E_OBJECT_CLASS_TYPE (object_class),
+				GTK_SIGNAL_OFFSET (ETreeClass, start_drag),
+				e_marshal_NONE__INT_POINTER_INT_POINTER,
+				GTK_TYPE_NONE, 4, GTK_TYPE_INT, GTK_TYPE_POINTER, GTK_TYPE_INT, GTK_TYPE_GDK_EVENT);
+
 	et_signals[TREE_DRAG_BEGIN] =
 		gtk_signal_new ("tree_drag_begin",
 				GTK_RUN_LAST,
-				object_class->type,
+				E_OBJECT_CLASS_TYPE (object_class),
 				GTK_SIGNAL_OFFSET (ETreeClass, tree_drag_begin),
 				e_marshal_NONE__INT_POINTER_INT_POINTER,
 				GTK_TYPE_NONE, 4,
@@ -2220,7 +2580,7 @@ e_tree_class_init (ETreeClass *class)
 	et_signals[TREE_DRAG_END] =
 		gtk_signal_new ("tree_drag_end",
 				GTK_RUN_LAST,
-				object_class->type,
+				E_OBJECT_CLASS_TYPE (object_class),
 				GTK_SIGNAL_OFFSET (ETreeClass, tree_drag_end),
 				e_marshal_NONE__INT_POINTER_INT_POINTER,
 				GTK_TYPE_NONE, 4,
@@ -2231,7 +2591,7 @@ e_tree_class_init (ETreeClass *class)
 	et_signals[TREE_DRAG_DATA_GET] =
 		gtk_signal_new ("tree_drag_data_get",
 				GTK_RUN_LAST,
-				object_class->type,
+				E_OBJECT_CLASS_TYPE (object_class),
 				GTK_SIGNAL_OFFSET (ETreeClass, tree_drag_data_get),
 				e_marshal_NONE__INT_POINTER_INT_POINTER_POINTER_UINT_UINT,
 				GTK_TYPE_NONE, 7,
@@ -2245,7 +2605,7 @@ e_tree_class_init (ETreeClass *class)
 	et_signals[TREE_DRAG_DATA_DELETE] =
 		gtk_signal_new ("tree_drag_data_delete",
 				GTK_RUN_LAST,
-				object_class->type,
+				E_OBJECT_CLASS_TYPE (object_class),
 				GTK_SIGNAL_OFFSET (ETreeClass, tree_drag_data_delete),
 				e_marshal_NONE__INT_POINTER_INT_POINTER,
 				GTK_TYPE_NONE, 4,
@@ -2257,7 +2617,7 @@ e_tree_class_init (ETreeClass *class)
 	et_signals[TREE_DRAG_LEAVE] =
 		gtk_signal_new ("tree_drag_leave",
 				GTK_RUN_LAST,
-				object_class->type,
+				E_OBJECT_CLASS_TYPE (object_class),
 				GTK_SIGNAL_OFFSET (ETreeClass, tree_drag_leave),
 				e_marshal_NONE__INT_POINTER_INT_POINTER_UINT,
 				GTK_TYPE_NONE, 5,
@@ -2269,7 +2629,7 @@ e_tree_class_init (ETreeClass *class)
 	et_signals[TREE_DRAG_MOTION] =
 		gtk_signal_new ("tree_drag_motion",
 				GTK_RUN_LAST,
-				object_class->type,
+				E_OBJECT_CLASS_TYPE (object_class),
 				GTK_SIGNAL_OFFSET (ETreeClass, tree_drag_motion),
 				e_marshal_BOOL__INT_POINTER_INT_POINTER_INT_INT_UINT,
 				GTK_TYPE_BOOL, 7,
@@ -2283,7 +2643,7 @@ e_tree_class_init (ETreeClass *class)
 	et_signals[TREE_DRAG_DROP] =
 		gtk_signal_new ("tree_drag_drop",
 				GTK_RUN_LAST,
-				object_class->type,
+				E_OBJECT_CLASS_TYPE (object_class),
 				GTK_SIGNAL_OFFSET (ETreeClass, tree_drag_drop),
 				e_marshal_BOOL__INT_POINTER_INT_POINTER_INT_INT_UINT,
 				GTK_TYPE_BOOL, 7,
@@ -2297,7 +2657,7 @@ e_tree_class_init (ETreeClass *class)
 	et_signals[TREE_DRAG_DATA_RECEIVED] =
 		gtk_signal_new ("tree_drag_data_received",
 				GTK_RUN_LAST,
-				object_class->type,
+				E_OBJECT_CLASS_TYPE (object_class),
 				GTK_SIGNAL_OFFSET (ETreeClass, tree_drag_data_received),
 				e_marshal_NONE__INT_POINTER_INT_POINTER_INT_INT_POINTER_UINT_UINT,
 				GTK_TYPE_NONE, 9,
@@ -2311,14 +2671,14 @@ e_tree_class_init (ETreeClass *class)
 				GTK_TYPE_UINT,
 				GTK_TYPE_UINT);
 
-	gtk_object_class_add_signals (object_class, et_signals, LAST_SIGNAL);
+	E_OBJECT_CLASS_ADD_SIGNALS (object_class, et_signals, LAST_SIGNAL);
 
 	class->set_scroll_adjustments = set_scroll_adjustments;
 
 	widget_class->set_scroll_adjustments_signal =
 		gtk_signal_new ("set_scroll_adjustments",
 				GTK_RUN_LAST,
-				object_class->type,
+				E_OBJECT_CLASS_TYPE (object_class),
 				GTK_SIGNAL_OFFSET (ETreeClass, set_scroll_adjustments),
 				gtk_marshal_NONE__POINTER_POINTER,
 				GTK_TYPE_NONE, 2, GTK_TYPE_ADJUSTMENT, GTK_TYPE_ADJUSTMENT);
@@ -2333,6 +2693,8 @@ e_tree_class_init (ETreeClass *class)
 				 GTK_ARG_WRITABLE, ARG_DRAW_FOCUS);
 	gtk_object_add_arg_type ("ETree::ETreeTableAdapter", GTK_TYPE_OBJECT,
 				 GTK_ARG_READABLE, ARG_ETTA);
+	gtk_object_add_arg_type ("ETree::uniform_row_height", GTK_TYPE_BOOL,
+				 GTK_ARG_READWRITE, ARG_UNIFORM_ROW_HEIGHT);
 }
 
 E_MAKE_TYPE(e_tree, "ETree", ETree, e_tree_class_init, e_tree_init, PARENT_TYPE);
